/*
* Distributed Computational System
* Client Module
*
* $Id: client.c,v 1.6 2001/11/29 07:06:17 kris Exp $
*/

#include "client_plugin.h"
#include "protocol.h"
#include "port.h"
#include <stdio.h>
#include <string.h>

#ifdef _WIN32
	#pragma comment(lib,"ws2_32.lib")
	#include <winsock2.h>
	#include <windows.h>
	#define PLUGIN_NAME "cp_primes.dll"
#else
	#include <netdb.h>
	#include <netinet/in.h>
	#include <sys/socket.h>
	#include <unistd.h>
	#include <dlfcn.h>
	#define PLUGIN_NAME "./cp_primes.so"
#endif

typedef struct _plugin_t {
#ifdef _WIN32
	HMODULE plug;
#else
	void *plug;
#endif
	cp_init_f cp_init;
	cp_getversion_f cp_getversion;
	cp_assignproblemchunk_f cp_assignproblemchunk;
} plugin_t;

int connect_to_server( const char *ip, unsigned short port, struct sockaddr_in *serv );
void fatal_error( const char *err );
int load_plugin( const char *file, plugin_t *plugin );

char *g_server_ip;
unsigned short g_server_port;
unsigned char g_cur_chunk_id[16];

int main( int argc, char **argv )
{
#ifdef _WIN32
	WSADATA wsadata;
#endif
	struct sockaddr_in serv;
	int sock;
	command_t command;
	client_chunk_t chunk;
	plugin_t plug;

	if( argc < 3 ) {
		fprintf( stderr, "Usage: %s <server ip> <server port>\n", argv[0] );
		return 1;
	}

#ifdef _WIN32
	if( WSAStartup( MAKEWORD(2,2), &wsadata ) != 0 ) {
		fprintf( stderr, "WSAStartup failed!\n" );
		return 1;
	}
#endif

	g_server_ip = argv[1];
	g_server_port = (unsigned short)atoi( argv[2] );

	if( load_plugin( PLUGIN_NAME, &plug ) < 0 )
		fatal_error( "Unable to load plugin" );
	(plug.cp_init)( 0 );

	/* enter our loop of fun! */
	while( 1 ) {
		if( (sock = connect_to_server( g_server_ip, g_server_port, &serv )) < 0 )
			fatal_error( "Unable to connect to server" );
		command.cmd = 12;
		command.len = 0;
		command.data = 0;
		if( PR_Send( sock, 0, &command ) < 0 )
			fatal_error( "Unable to send data to server" );
		if( PR_RecvBlocking( sock, &command ) != 2 )
			fatal_error( "Error receiving reply" );
		close( sock );
		if( command.cmd != 10 || command.len < 24 ) {
			FREE_COMMAND_DATA( command );
			fatal_error( "Unexpected command or data" );
		}

		memcpy( g_cur_chunk_id, command.data, 16 );
		memcpy( &chunk.len, command.data + 20, 4 );
		if( command.len < (24 + chunk.len) ) {
			FREE_COMMAND_DATA( command );
			fatal_error( "Invalid packet" );
		}
		chunk.data = command.data + 24;
		(plug.cp_assignproblemchunk)( CL_ReportResult, 0, &chunk );
		FREE_COMMAND_DATA( command );
	}

	return 0;
}

int CL_ReportResult( int plugin_id, client_result_t *result )
{
	int sock;
	struct sockaddr_in serv;
	command_t command;

	/*
		THIS SUCKS.  We have to allocate a whole new
		buffer to tack on extra info into the data part
		of the packet.  This could be solved fairly
		simply with a PR_Send function that took
		varargs, or that took some sort of fun struct
		and sent each item in order (so no new
		allocation would be required).  The struct idea
		could also easily support the passing of file
		handles of some sort, to completely eliminate
		memory allocation for result data altogether.
	*/
	command.cmd = 20;
	command.len = result->len + 21;
	command.data = (unsigned char *)malloc( command.len );
	memcpy( command.data, g_cur_chunk_id, 16 );
	command.data[16] = 0;
	memcpy( command.data + 17, &result->len, 4 );
	memcpy( command.data + 21, result->data, result->len );

	if( (sock = connect_to_server( g_server_ip, g_server_port, &serv )) < 0 )
		fatal_error( "Unable to connect to server" );
	if( PR_Send( sock, 0, &command ) < 0 ) {
		FREE_COMMAND_DATA( command );
		fatal_error( "Unable to send data to server" );
	}
	FREE_COMMAND_DATA( command );
	if( PR_RecvBlocking( sock, &command ) != 2 )
		fatal_error( "Error receiving reply" );
	close( sock );
	if( command.cmd != 20 ) { /* 20 == Acknowledge */
		FREE_COMMAND_DATA( command );
		fatal_error( "Unexpected response" );
	}
	FREE_COMMAND_DATA( command );

	return 0;
}

int connect_to_server( const char *ip, unsigned short port, struct sockaddr_in *serv )
{
	int fd = -1;
	struct hostent *he;

	if( (fd = socket( AF_INET, SOCK_STREAM, 0 )) < 0 )
		return -1;
	memset( serv, 0, sizeof( struct sockaddr_in ) );
	serv->sin_family = AF_INET;
	serv->sin_port = htons( port );
	he = (struct hostent *)gethostbyname( ip );
	if( he == NULL ) {  
		unsigned long addr;
		if( (addr = inet_addr( ip )) == INADDR_ERR ) {
			close( fd );
			return -1;
		}
		memcpy( &serv->sin_addr, &addr, 4 );
	} else
		serv->sin_addr = *((struct in_addr *)he->h_addr);

	if( connect( fd, (struct sockaddr *)serv, sizeof( struct sockaddr_in ) ) < 0 ) {
		close( fd );
		return -1;   
	}
	return fd;
}

void fatal_error( const char *err )
{
	fprintf( stderr, "%s\n", err );
	exit( 1 );
}

#ifdef _WIN32
int load_plugin( const char *file, plugin_t *plugin )
{
	plugin->plug = (HMODULE)LoadLibrary( file );
	if( !plugin->plug )
		return -1;
	if( !(plugin->cp_init = (cp_init_f)GetProcAddress( plugin->plug, "CP_Init" )) )
		return -1;
	if( !(plugin->cp_getversion = (cp_getversion_f)GetProcAddress( plugin->plug, "CP_GetVersion" )) )
		return -1;
	if( !(plugin->cp_assignproblemchunk =
		(cp_assignproblemchunk_f)GetProcAddress( plugin->plug, "CP_AssignProblemChunk" )) )
		return -1;
	return 0;
}
#else
int load_plugin( const char *file, plugin_t *plugin )
{
	char *error;

	plugin->plug = dlopen( file, RTLD_LAZY );
	if( !plugin->plug )
		return -1;
	dlerror(); /* clear error */
	plugin->cp_init = dlsym( plugin->plug, "CP_Init" );
	if( (error = dlerror()) != NULL )
		return -1;
	plugin->cp_getversion = dlsym( plugin->plug, "CP_GetVersion" );
	if( (error = dlerror()) != NULL )
		return -1;
	plugin->cp_assignproblemchunk = dlsym( plugin->plug, "CP_AssignProblemChunk" );
	if( (error = dlerror()) != NULL )
		return -1;
	return 0;
}
#endif

